/*
 * Copyright (C) 1996-2003 LAAS/CNRS 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include "config.h"
__RCSID("$LAAS: request.c,v 1.19 2005/10/27 12:54:33 matthieu Exp $");

#include <stdio.h>
#include <math.h>
#include <tcl.h>
#include <stdlib.h>
#include <string.h>

#include "gdhe/GDHE_packets.h"

#include "global.h"
#include "request.h"

/**
 ** Interpretation et execution d'un paquet GDHE 
 **/
static void place_robot(int which, position *pos);
static void select_robot_type(int which, int type);
static void erase_robot(int which);
static void place_walls(int which, v_segment *vseg);
static void erase_walls(int which);
static void place_robot_traj(int which, trajectory *traj);
static void erase_robot_traj(int which);
static void place_rs_traj(int which, rs_trajectory *rs_traj);
static void erase_rs_traj(int which);
static void update_per_obst(polygon *pol);
static void erase_per_obst(void);
static void place_dependency_edge(int from, int to);
static void erase_dependency_edge(int from, int to);
static void place_dependency_cycle(int which, dependency_cycle *cycle);
static void erase_dependency_cycle(int which);
#ifdef HAVE_GDHE_TERRAIN
static void place_terrain(int which, GDHE_MNT *terrain);
static void erase_terrain(int which);
#endif
static int exist_robot(char *which);


void
request_exec(packet *data)
{
    static char* eval_buffer = 0;
    static int   eval_buffer_length = 0;
    static int   eval_buffer_index = 0;
    char *new_buffer;

    int which;
    
    if (data == NULL) {
        return;
    }
    which = data->which;
    /* printf("request_exec %d %d\n", data->which, data->command);  */

    switch (data->command) {
	
      case PLACE_ROBOT:
	place_robot(which, data->args.pos);
	break;
	
      case ERASE_ROBOT:
	erase_robot(which);
	break;
	
      case PLACE_WALLS:
	place_walls(which, data->args.vseg);
	break;
	
      case ERASE_WALLS:
	erase_walls(which);
	break;
	
      case PLACE_ROBOT_TRAJ:
	place_robot_traj(which, data->args.traj);
	break;
	
      case ERASE_ROBOT_TRAJ:
	erase_robot_traj(which);
	break;
	
      case PLACE_RS_TRAJ:
	place_rs_traj(which, data->args.rs_traj);
	break;
	
      case ERASE_RS_TRAJ:
	erase_rs_traj(which);
	break;
	
      case SELECT_ROBOT_TYPE:
	select_robot_type(which, *data->args.robot_type);
	break;
	
      case UPDATE_PER_OBST:
	update_per_obst(data->args.pol);
	break;
	
      case ERASE_PER_OBST:
	erase_per_obst();
	break;
	
      case PLACE_DEPENDENCY_EDGE:
	place_dependency_edge(which, *data->args.edge_to);
	break;
	
      case ERASE_DEPENDENCY_EDGE:
	erase_dependency_edge(which, *data->args.edge_to);
	break;

      case PLACE_DEPENDENCY_CYCLE:
	place_dependency_cycle(which, data->args.cycle);
	break;

      case ERASE_DEPENDENCY_CYCLE:
	erase_dependency_cycle(which);
	break;

      case PLACE_MISSION_GOAL:
	break;
      case ERASE_MISSION_GOAL:
	break;
	
      case DISCONNECT:
	break;
	
      case CLOSE_ALL:
	break;
	
      case EVAL_EXPRESSION_ADD:
      case EVAL_EXPRESSION:
        if (! eval_buffer_length)
        {
            eval_buffer = (char*)malloc(EVAL_PACKET_LENGTH);
            eval_buffer_length = EVAL_PACKET_LENGTH;
        }
        else if (eval_buffer_index + EVAL_PACKET_LENGTH > eval_buffer_length)
        {
            size_t new_length = eval_buffer_length + EVAL_PACKET_LENGTH;
            if (new_length >= EVAL_BUFFER_LENGTH)
            {
                eval_buffer_index = 0;
                return;
            }
            
            new_buffer = (char*)malloc(new_length);
            memcpy(new_buffer, eval_buffer, eval_buffer_length);

            free(eval_buffer);
            eval_buffer        = new_buffer;
            eval_buffer_length = new_length;
        }
        strncpy(&eval_buffer[eval_buffer_index], data->args.str, EVAL_PACKET_LENGTH - 1);
        eval_buffer_index += strlen(data->args.str);
        eval_buffer[eval_buffer_index] = 0;

        if (data->command == EVAL_EXPRESSION)
        {
            if (Tcl_Eval(tclInterp, eval_buffer) == TCL_ERROR) {
                Tcl_BackgroundError(tclInterp);
            }
            eval_buffer_index = 0;
        }
	break;

#ifdef HAVE_GDHE_TERRAIN
      case PLACE_TERRAIN:
	place_terrain(which, data->args.terrain); 
	break;
	
      case ERASE_TERRAIN:
	erase_terrain(which); 
	break;
#endif /* HAVE_GDHE_TERRAIN */
	
	/* Commandes non traite'es par la version X */   
      case PLACE_TRAJECTORY:
	fprintf(stderr,"traj \n");
	break;
      case ERASE_TRAJECTORY:
	fprintf(stderr,"erase traj\n");
	break;

      case SET_PLATFORM:
      case PLACE_DIAMOND:
      case ERASE_DIAMOND:
      case PLACE_DOORS:
      case ERASE_DOORS:
      case PLACE_WALLS_REL:
      case PLACE_WALLS_ACC:
      case PLACE_FREE_SPACE:
      case ERASE_FREE_SPACE:

      case PLACE_PATH:
      case ERASE_PATH:
      case PLACE_ROBOT_PATH:
      case ERASE_ROBOT_PATH:
      case PLACE_GOAL_O:
      case PLACE_GOAL_NO:
      case ERASE_GOAL:
      case EXEC_TRAJ:
      case MOVE_PLATFORM:
      case VIDEO_ON:
      case VIDEO_OFF:
      case LASER_ON:
      case LASER_OFF:
	
	break;
	
      default:
	fprintf(stderr, "GDHE: unknown command %d\n", data->command);
    } /* switch */

} /* request_exec */

/*----------------------------------------------------------------------*/

static void
place_robot(int which, position *pos)
{
    char robot[14];
    char buf[1024];

    snprintf(robot, sizeof(robot), "r%d", which);
    /* Test de la validite' de which */
    if (!exist_robot(robot)) {
	return;
    }
    snprintf(buf, sizeof(buf), "%f", pos->p->x);
    Tcl_SetVar2(tclInterp, "pos", robot, buf, 
		TCL_LIST_ELEMENT |TCL_GLOBAL_ONLY);
    snprintf(buf, sizeof(buf), "%f", pos->p->y);
    Tcl_SetVar2(tclInterp, "pos", robot, buf, 
		TCL_APPEND_VALUE|TCL_LIST_ELEMENT |TCL_GLOBAL_ONLY);
    snprintf(buf, sizeof(buf), "%f", RAD_TO_DEG(pos->theta));
    Tcl_SetVar2(tclInterp, "pos", robot, buf, 
		TCL_APPEND_VALUE|TCL_LIST_ELEMENT |TCL_GLOBAL_ONLY);



} /* place_robot */


/*----------------------------------------------------------------------*/

static void
select_robot_type(int which, int type)
{
    char buf[1024];

    /* Test de la validite' de which */
    if (which < 0 || which > MAX_ROBOT) {
	fprintf(stderr, "GDHE: paquet vrol: which: %d\n", which);
	return;
    }
    snprintf(buf, sizeof(buf), "set robots(r%d) $robotType(%d)", which, type);
    Tcl_Eval(tclInterp, buf);
    snprintf(buf, sizeof(buf), "set platform(r%d) $platformType(%d)", 
	which, type);
    Tcl_Eval(tclInterp, buf);
    snprintf(buf, sizeof(buf), "eval etat_pmo %d 0", which);
    Tcl_Eval(tclInterp, buf);

} /* select_robot_type */

/*----------------------------------------------------------------------*/

static void 
erase_robot(int which)
{
    char robot[14];

    snprintf(robot, sizeof(robot), "r%d", which);
    /* Test de la validite' de which */
    if (!exist_robot(robot)) {
	return;
    }
    Tcl_UnsetVar2(tclInterp, "pos", robot, TCL_GLOBAL_ONLY);
    

} /* erase_robot */

/*----------------------------------------------------------------------*/

void
place_walls(int which, v_segment *vseg)
{
    int id, i, n;
    char robot[20];
    char *buf[4];
    char *p;
    
    id = which % PACKET_DIVIDER;
    which = which / PACKET_DIVIDER;
    snprintf(robot, sizeof(robot), "r%d", which);

    /* Test de la validite' de which */
    if (!exist_robot(robot)) {
	return;
    }
    snprintf(robot, sizeof(robot), "r%d,%d", which, id);
    /* Efface anciens murs */
    p = Tcl_GetVar2(tclInterp, "walls", robot, TCL_GLOBAL_ONLY);
    if (p != NULL) {
	Tcl_UnsetVar2(tclInterp, "walls", robot, TCL_GLOBAL_ONLY);
    }
	
    n = vseg->number;
    for (i = 0; i < 4; i++) {
	buf[i] = Tcl_Alloc(20);
    }
    
    for (i = 0; i < n; i++) {
	snprintf(buf[0], 20, "%f", vseg->array[i].p1->x);
	snprintf(buf[1], 20, "%f", vseg->array[i].p1->y);
	snprintf(buf[2], 20, "%f", vseg->array[i].p2->x);
	snprintf(buf[3], 20, "%f", vseg->array[i].p2->y);
	p = Tcl_Merge(4, buf);

	Tcl_SetVar2(tclInterp, "walls", robot, p, 
		    TCL_APPEND_VALUE | TCL_LIST_ELEMENT | TCL_GLOBAL_ONLY);
	Tcl_Free(p);
	
    } /* for */
    
    for (i = 0 ; i < 4; i++) {
	Tcl_Free(buf[i]);
    }

} /* place_walls */

/*----------------------------------------------------------------------*/

static void
erase_walls(int which)
{
    int id;
    char robot[20];

    char *p;
    
    id = which % PACKET_DIVIDER;
    which = which / PACKET_DIVIDER;
    snprintf(robot, sizeof(robot), "r%d", which);
	
    /* Test de la validite' de which */
    if (!exist_robot(robot)) {
	return;
    }
    snprintf(robot, sizeof(robot), "r%d,%d", which, id);
    /* Efface anciens murs */
    p = Tcl_GetVar2(tclInterp, "walls", robot, TCL_GLOBAL_ONLY);
    if (p != NULL) {
	Tcl_UnsetVar2(tclInterp, "walls", robot, TCL_GLOBAL_ONLY);
    }
    
    
} /* erase_walls */

/*----------------------------------------------------------------------*/

static void
place_robot_traj(int which, trajectory *traj)
{
    char robot[20];
    char buf[1024];
    int i;

    snprintf(robot, sizeof(robot), "r%d", which);
    /* Test de la validite' de which */
    if (!exist_robot(robot)) {
	return;
    }
    snprintf(buf, sizeof(buf), "%d", traj->vloc->number);
    Tcl_SetVar2(tclInterp, "trajectory", robot, buf, 
		TCL_LIST_ELEMENT |TCL_GLOBAL_ONLY);
    for (i=0;i<traj->vloc->number;i++) {
      
        snprintf(buf, sizeof(buf), "%f", traj->vloc->array[i].x);
	Tcl_SetVar2(tclInterp, "trajectory", robot, buf, 
		TCL_APPEND_VALUE|TCL_LIST_ELEMENT |TCL_GLOBAL_ONLY);
	snprintf(buf, sizeof(buf), "%f", traj->vloc->array[i].y);
	Tcl_SetVar2(tclInterp, "trajectory", robot, buf, 
		TCL_APPEND_VALUE|TCL_LIST_ELEMENT |TCL_GLOBAL_ONLY);
	Tcl_SetVar2(tclInterp, "trajectory", robot, "0", 
		TCL_APPEND_VALUE|TCL_LIST_ELEMENT |TCL_GLOBAL_ONLY);
    }
	
  
} /* place_robot_traj */

/*----------------------------------------------------------------------*/

static void 
erase_robot_traj(int which)
{
    char robot[20];
    char *p;
  
    snprintf(robot, sizeof(robot), "r%d", which);
    /* Test de la validite' de which */
    if (!exist_robot(robot)) {
	return;
    }
    /* Efface anciens trajectoire robot */
    p=Tcl_GetVar2(tclInterp,"trajectory",robot,TCL_GLOBAL_ONLY);
    if (p != NULL ) {
      Tcl_UnsetVar2(tclInterp,"trajectory",robot,TCL_GLOBAL_ONLY);
    } 
	
  
} /* erase_robot_traj */


/*----------------------------------------------------------------------*/

static void
place_rs_traj(int which, rs_trajectory *rs_traj)
{     
    char robot[20];
    char robotID[20];
    char buf[1024];
    char *p;
    int i,id;

    id=which%PACKET_DIVIDER;
    which=which/PACKET_DIVIDER;

    snprintf(robotID, sizeof(robotID), "r%d%d", which,id);
    snprintf(robot, sizeof(robot), "r%d", which);

    /* Test de la validite' de which */
    if (!exist_robot(robot)) {
	return;
    }

    p=Tcl_GetVar2(tclInterp,"rs_trajectory",robot,TCL_GLOBAL_ONLY);
    if (p == NULL ) {
      snprintf(buf, sizeof(buf), "%d", 1);
      Tcl_SetVar2(tclInterp,"rs_trajectory",robot,buf,
		  TCL_LIST_ELEMENT|TCL_GLOBAL_ONLY);
    }

    snprintf(buf, sizeof(buf), "%f", rs_traj->initpos->p->x);
    Tcl_SetVar2(tclInterp, "rs_trajectory", robotID, buf, 
		TCL_LIST_ELEMENT |TCL_GLOBAL_ONLY);
    snprintf(buf, sizeof(buf), "%f", rs_traj->initpos->p->y);
    Tcl_SetVar2(tclInterp, "rs_trajectory", robotID, buf, 
		TCL_APPEND_VALUE|TCL_LIST_ELEMENT |TCL_GLOBAL_ONLY);
    snprintf(buf, sizeof(buf), "%f", rs_traj->initpos->theta);
    Tcl_SetVar2(tclInterp, "rs_trajectory", robotID, buf, 
		TCL_APPEND_VALUE|TCL_LIST_ELEMENT |TCL_GLOBAL_ONLY);

    snprintf(buf, sizeof(buf), "%d", rs_traj->number);
    Tcl_SetVar2(tclInterp, "rs_trajectory", robotID, buf, 
		TCL_APPEND_VALUE|TCL_LIST_ELEMENT |TCL_GLOBAL_ONLY);
   
    for (i=0;i<rs_traj->number;i++) {      
        snprintf(buf, sizeof(buf), "%f", rs_traj->array[i].dt);
	Tcl_SetVar2(tclInterp, "rs_trajectory", robotID, buf, 
		TCL_APPEND_VALUE|TCL_LIST_ELEMENT |TCL_GLOBAL_ONLY);
	snprintf(buf, sizeof(buf), "%f", rs_traj->array[i].dl);
	Tcl_SetVar2(tclInterp, "rs_trajectory", robotID, buf, 
		TCL_APPEND_VALUE|TCL_LIST_ELEMENT |TCL_GLOBAL_ONLY);
    }
	
  
} /* place_rs_traj */


/*----------------------------------------------------------------------*/

static void
erase_rs_traj(int which)
{
    char robot[20];
    char robotID[20];
    char *p;
    int id;
  
    which=which/PACKET_DIVIDER;
    id=which%PACKET_DIVIDER;

    snprintf(robotID, sizeof(robotID), "r%d%d", which,id);
    snprintf(robot, sizeof(robot), "r%d",which);

    /* Test de la validite' de which */
    if (!exist_robot(robot)) {
	return;
    }
    /* Efface anciens rs_trajectoire */
    p=Tcl_GetVar2(tclInterp,"rs_trajectory",robotID,TCL_GLOBAL_ONLY);
    if (p != NULL ) {
      Tcl_UnsetVar2(tclInterp,"rs_trajectory",robotID,TCL_GLOBAL_ONLY);
    }
    
} /* erase_rs_traj */

/*----------------------------------------------------------------------*/

 
static void
update_per_obst(polygon *pol)
{
    int i, j, n, m;
    char **vertex, *p;
  

    Tcl_UnsetVar(tclInterp, "perObst", TCL_GLOBAL_ONLY);
    
    n = pol->number;
    
    for (i = 0; i < n; i++) {
	m = pol->polys[i].number;
	vertex = (char **)Tcl_Alloc(sizeof(char *)*m*2);
	for (j = 0; j < m; j++) {
	    vertex[2*j] = Tcl_Alloc(20);
	    vertex[2*j+1] = Tcl_Alloc(20);
	    snprintf(vertex[2*j], 20, "%f", pol->polys[i].array[j].x);
	    snprintf(vertex[2*j+1], 20, "%f", pol->polys[i].array[j].y);
	} /* for */
	p = Tcl_Merge(2*m, vertex);
	/* printf("vertex %s\n", p); */
	Tcl_SetVar(tclInterp, "perObst", p, 
		   TCL_APPEND_VALUE | TCL_LIST_ELEMENT | TCL_GLOBAL_ONLY);
	Tcl_Free(p);
	for (j = 0; j < 2*m; j++) {
	    Tcl_Free(vertex[j]);
	}
	Tcl_Free((char *)vertex);
    } /* for */

} /* update_per_obst */

/*----------------------------------------------------------------------*/

static void
place_dependency_edge(int from, int to)
{
} /* place_dependency_edge */

/*----------------------------------------------------------------------*/

static void
erase_dependency_edge(int from, int to)
{
} /* erase_dependency_edge */

/*----------------------------------------------------------------------*/

static void
place_dependency_cycle(int which, dependency_cycle *cycle)
{
} /* place_dependency_cycle */

/*----------------------------------------------------------------------*/

static void
erase_dependency_cycle(int which)
{
} /* erase_dependency_cycle */

/*----------------------------------------------------------------------*/

static void
erase_per_obst(void)
{
    Tcl_UnsetVar(tclInterp, "perObst", TCL_GLOBAL_ONLY);

} /* erase_per_obst */

/*----------------------------------------------------------------------*/

#ifdef HAVE_GDHE_TERRAIN

static void
place_terrain(int which, GDHE_MNT *terrain)
{
    char buf[20];

    snprintf(buf, sizeof(buf), "terrain%d", which);
    /* defineTerrain(buf, terrain); */
    Tcl_SetVar(tclInterp, "mnt", buf, TCL_GLOBAL_ONLY | TCL_APPEND_VALUE);
    
}

/*----------------------------------------------------------------------*/

static void
erase_terrain(int which)
{
    char buf[20], cmd[1024];

    snprintf(buf, sizeof(buf), "terrain%d", which);
    snprintf(cmd, sizeof(cmd),
	    "set i [lsearch -exact %s mnt]\n"
	    "if { $i > 0} { set mnt [lreplace $mnt $i $i] }\n", buf);
    Tcl_Eval(tclInterp, cmd);
    /* deleteTerrain(buf); */
}
#endif /* HAVE_GDHE_TERRAIN */

/*----------------------------------------------------------------------*/

/**
 ** Test de la validite' de which 
 **/
static int
exist_robot(char *which)
{
    if (Tcl_GetVar2(tclInterp, "robots", which, TCL_GLOBAL_ONLY) != NULL) {
	return 1;
    } else {
	return 0;
    }
    
}
