# 
# Copyright (C) 1996-2001 LAAS/CNRS 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    - Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# 
# $LAAS: karma.tcl,v 1.5 2006/08/02 13:05:17 sjoyeux Exp $
#----------------------------------------------------------------------
package provide Gdhe::models 1.0

# Karma (made by Simon in 12/2002)

#----------------------------------------------------------------------
#
# karma
#
# - vectorization : the vectorization angle of the propellers (degrees)
# - displayFOV : display the front cameras FOV (0, 1 or 2 - see cameraWithFOV)
# - frontFrame : display the front camera frame (0 or 1)
# - depthFOV : the depth of the lines (or planes) that defines the FOV, unused if displayFOV is 0
#
#----------------------------------------------------------------------

proc karma { { vectorization 0 } { frontFrame 1 } { displayFOV 0 } { depthFOV 0 } } {

    # The gondola (a rectangle blue box)
    color 100 100 255
    box 0 0 -0.125 1.0 0.25 0.25

    pushMatrix
    color 0 0 255
    rotate 180 1 0 0 
    repere 1.0
    popMatrix

    # The vectorized propellers
    karmaPropellers $vectorization

    # The stereo bench
    karmaStereoBench $displayFOV $frontFrame $depthFOV

    # The enveloppe
    pushMatrix
    translate 3.0 0.0 1.1
    rotate 180 0 0 1
    enveloppe

    popMatrix
}

# The stereo bench
proc karmaStereoBench {displayFOV frontFrame depthFOV} {

    set hFOV 33.5
    set vFOV 26.5

    color 0 0 0 
    cylinder -1.2 0 0 x 0.03 2.4;	# The axis

    pushMatrix
    translate 1.2 0 0
    rotate 90 0 1 0
    rotate 90 1 0 0
    cameraWithFOV $hFOV $vFOV $displayFOV $depthFOV $frontFrame
    popMatrix
}


# The vectorized propellers
proc karmaPropellers {vectorization} {
    pushMatrix

    translate -0.25 0 0;		# 1/4 of the gondola backward
    rotate $vectorization 0 1 0
    color 0 0 0 
    cylinder 0 -0.5 0 y 0.03 1.0;	# The axis

    pushMatrix
    translate 0 0.5 0 
    color 150 150 150
    cylinder 0 0 0 x 0.3 0.03;		# Left propeller
    popMatrix

    pushMatrix
    translate 0 -0.5 0 
    color 150 150 150
    cylinder 0 0 0 x 0.3 0.03;		# Right propeller
    popMatrix

    popMatrix
}

#----------------------------------------------------------------------
#
# karma internal control parameters
#
#----------------------------------------------------------------------

proc karmaControls {} {
    global frontFrame
    global frontFOV
    global depthFOV

    toplevel .karma
    wm title .karma "Karma controls"
    set dpyWindow .gdhe.ogl

#     wm geometry .karma -100-100

    # The internal configuration parameters
    scale  .karma.vectorization -length 200  -label {Vectorization} -from 80 -to 180 \
	-resolution 1.0 -command {setKarma vectorization } -orient horizontal
    pack .karma.vectorization -fill x

    # The front camera
    # TODO : mettre une barre de sparation
    frame .karma.frontCamera 
    pack .karma.frontCamera -fill x

    label .karma.frontCamera.label -text "Front camera"
    pack .karma.frontCamera.label -fill x

    # Toggle the frame display
    checkbutton .karma.frontCamera.frame -text "Display frame" \
	-variable frontFrame \
    	-command "$dpyWindow redraw" 
    pack .karma.frontCamera.frame -fill x

    # Toggle the field of view display
    set displayFOVmenu \
	[tk_optionMenu .karma.frontCamera.fov displayFOV "No field of view" \
	     "Line FOV" "Filled FOV"]
    .karma.frontCamera.fov configure -width 20
    $displayFOVmenu entryconfigure 0 \
	-command "set displayFOV 0; $dpyWindow redraw"
    $displayFOVmenu entryconfigure 1 \
	-command "set displayFOV 1; $dpyWindow redraw"
    $displayFOVmenu entryconfigure 2 \
	-command "set displayFOV 2; $dpyWindow redraw"
    pack .karma.frontCamera.fov -fill x

    # Changes the the field of view depth
    set depthFOV 10.0;			# TODO : why is this necessary ?
    scale  .karma.frontCamera.depth -length 200  -label {FOV depth (m)} \
	-from 0.0 -to 100.0 -resolution 1.0 \
	-command {setKarma frontCamera.depth } -orient horizontal
    .karma.frontCamera.depth set 10
    pack .karma.frontCamera.depth -fill x

}

# Front camera field of view
# proc setFrontFOV {window} {
#     global pipoFrontFOV
#     global displayFOV

#     set displayFOV $pipoFrontFOV
# }

# Slider events handler
proc setKarma { axis value } {
    global robots
    global depthFOV

    set depthFOV [.karma.frontCamera.depth get]

    set robots(karma) { karma [.karma.vectorization get] }


    .gdhe.ogl redraw
}

### 
### A camera with its field of view
###
proc cameraWithFOV {hFOV vFOV displayFOV depthFOV frontFrame} {
    # hFOV and vFOV : horizontal and vertical half field of view in degrees 
    # displayFOV :	0 -> no display
    #			1 -> display lines
    #			2 -> display planes
    # depthFOV : the depth of the lines (or planes) that defines the FOV
    # frontFrame : draw the camera frame if non 0

    # Camera body definition
    object camera {
	color 120 120 120
	box -0.03 0.0 -0.02 0.06 0.04 0.04
	color 70 70 70
	cylinder 0.0 0.0 0.0 x 0.03 0.01
	cylinder 0.01 0.0 0.0 x 0.03 0.05 0.02
	cylinder 0.03 0.0 0.0 x 0.05 0.01
    }

    # Frame
    if {$frontFrame} {
	pushMatrix
	translate 0 0 0
	color 255 0 0 
	repere 1.0
	popMatrix
    }

    # Field of view
    if {$displayFOV != 0} {
	color 0 100 0

	set hFOVrad [expr $hFOV * 3.14159 / 180.0]
	set vFOVrad [expr $vFOV * 3.14159 / 180.0]
	set x $depthFOV
	set y [expr $depthFOV * tan($hFOVrad)]
	set z [expr $depthFOV * tan($vFOVrad)]

	# First display the lines
	# "top left" line
	polyline 2 0 0 0 $x $y $z
	# "top right" line
	polyline 2 0 0 0 $x [expr - $y] $z
	# "down right" line
	polyline 2 0 0 0 $x [expr - $y] [expr -$z]
	# "down left" line
	polyline 2 0 0 0 $x $y [expr -$z]
	
	# Down fronto-parallel line
	polyline 2 $x [expr - $y] [expr -$z] $x $y [expr -$z]
	# Top fronto-parallel line
	polyline 2 $x $y $z $x [expr - $y] $z
	# Left fronto-parallel line
	polyline 2 $x $y $z $x $y [expr -$z]
	# Right fronto-parallel line
	polyline 2 $x [expr - $y] $z $x [expr - $y] [expr -$z]

	if {$displayFOV == 2} {		# Planes are also displayed
	    cullFace false
	    color 0 255 0
	    # top plane
	    polygon 3 0 0 0 $x [expr - $y] $z  $x $y $z
	    # bottom plane
	    polygon 3 0 0 0 $x [expr - $y] [expr -$z] $x $y [expr -$z]
	    # left plane
	    polygon 3 0 0 0 $x $y $z $x $y [expr -$z]
	    # right plane
	    polygon 3 0 0 0 $x [expr - $y] $z $x [expr - $y] [expr -$z]
	}
    }
    
}

proc enveloppe {} {
    concave false
    cullFace True
    set nbCylinderFacets 30

# Airship enveloppe
    color 150 150 150
    # Front part (3.0m long wrt the largest diameter)
    cylinder 0 0 0 x -0 -0.34 0.05 $nbCylinderFacets
    cylinder 0.05 0 0 x -0.34 -0.48 0.05 $nbCylinderFacets
    cylinder 0.10 0 0 x -0.49 -0.60 0.06 $nbCylinderFacets
    cylinder 0.16 0 0 x -0.60 -0.76 0.09 $nbCylinderFacets
    cylinder 0.25 0 0 x -0.76 -1.06 0.25 $nbCylinderFacets
    cylinder 0.50 0 0 x -1.06 -1.26 0.25 $nbCylinderFacets
    cylinder 0.75 0 0 x -1.26 -1.42 0.25 $nbCylinderFacets
    cylinder 1.00 0 0 x -1.42 -1.56 0.25 $nbCylinderFacets
    cylinder 1.25 0 0 x -1.56 -1.66 0.25 $nbCylinderFacets
    cylinder 1.50 0 0 x -1.66 -1.74 0.25 $nbCylinderFacets
    cylinder 1.75 0 0 x -1.74 -1.81 0.25 $nbCylinderFacets
    cylinder 2.00 0 0 x -1.81 -1.86 0.25 $nbCylinderFacets
    cylinder 2.25 0 0 x -1.86 -1.89 0.25 $nbCylinderFacets
    cylinder 2.50 0 0 x -1.89 -1.91 0.25 $nbCylinderFacets
    cylinder 2.75 0 0 x -1.91 -1.92 0.25 $nbCylinderFacets

    # Rear part (6.4m long wrt the largest diameter)
    cylinder 3.00 0 0 x -1.92 -1.91 0.53 $nbCylinderFacets
    cylinder 3.53 0 0 x -1.91 -1.89 0.53 $nbCylinderFacets
    cylinder 4.06 0 0 x -1.89 -1.86 0.53 $nbCylinderFacets
    cylinder 4.59 0 0 x -1.86 -1.81 0.53 $nbCylinderFacets
    cylinder 5.12 0 0 x -1.81 -1.74 0.53 $nbCylinderFacets
    cylinder 5.65 0 0 x -1.74 -1.66 0.53 $nbCylinderFacets
    cylinder 6.18 0 0 x -1.66 -1.56 0.53 $nbCylinderFacets
    cylinder 6.71 0 0 x -1.56 -1.43 0.53 $nbCylinderFacets
    cylinder 7.24 0 0 x -1.43 -1.27 0.53 $nbCylinderFacets
    cylinder 7.77 0 0 x -1.27 -1.06 0.53 $nbCylinderFacets
    cylinder 8.30 0 0 x -1.06 -0.77 0.53 $nbCylinderFacets
    cylinder 8.83 0 0 x -0.77 -0.66 0.14 $nbCylinderFacets
    cylinder 8.97 0 0 x -0.66 -0.54 0.14 $nbCylinderFacets
    cylinder 9.11 0 0 x -0.54 -0.40 0.14 $nbCylinderFacets
    cylinder 9.24 0 0 x -0.40 -0.00 0.14 $nbCylinderFacets

#     cylinder 0 0 0 x 0 1 0.25
#     cylinder 0.25 0 0 x 1 1.5 0.5
#     cylinder 0.75 0 0 x 1.5 2 1.25
    
#     cylinder 2 0 0 x 2 2 1.5
 
#     cylinder 3.5 0 0 x 2 1.25 2.5
#     cylinder 6 0 0 x 1.25 0.5 1.75
#     cylinder 7.75 0 0 x 0.5 0 0.25

#Ailerons (vue de derriere)
    #haut droit
    color 70 70 255
    pushMatrix
    translate 8.4 0.5 0.5
    rotate 45 1 0 0 
    box 0 0 0 0.8 0.7 0.05 
    translate -0.45 -0.14 0
    rotate 40 0 0 1 
    box 0 0 0 0.68 0.69 0.05 
    popMatrix

    #haut gauche
    color 70 70 255
    pushMatrix
    translate 8.4 -0.5 0.5
    rotate 135 1 0 0 
    box 0 0 0 0.8 0.7 0.05 
    translate -0.45 -0.14 0
    rotate 40 0 0 1 
    box 0 0 0 0.68 0.69 0.05
    popMatrix

    #bas droit
    color 70 70 255
    pushMatrix
    translate 8.4 0.5 -0.5
    rotate -45 1 0 0 
    box 0 0 0 0.8 0.7 0.05 
    translate -0.45 -0.14 0
    rotate 40 0 0 1 
    box 0 0 0 0.68 0.69 0.05
    popMatrix
    #bas gauche
    pushMatrix
    translate 8.4 -0.5 -0.5
    rotate -135 1 0 0 
    box 0 0 0 0.8 0.7 0.05 
    translate -0.45 -0.14 0
    rotate 40 0 0 1 
    box 0 0 0 0.68 0.69 0.05 
    popMatrix

#Surface mobiles

    #haut droit
    pushMatrix
    translate 8.9 0.5 0.5
    rotate 45 1 0 0 
    color 70 70 255
    box 0 0 0 0.18 0.7 0.05
    popMatrix    

    #haut gauche
    pushMatrix
    translate 8.9 -0.5 0.5
    rotate 135 1 0 0 
    color 70 70 255
    box 0 0 0 0.18 0.7 0.05 
    popMatrix    

    #bas droit
    pushMatrix
    translate 8.9 0.5 -0.5
    rotate -45 1 0 0 
    color 70 70 255
    box 0 0 0 0.18 0.7 0.05 
    popMatrix    

    #bas gauche
    pushMatrix
    translate 8.9 -0.5 -0.5
    rotate -135 1 0 0
    color 70 70 255
    box 0 0 0 0.18 0.7 0.05 
    popMatrix    
}
